"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class CortexException(Exception):
    """
    Base exception type.
    """
    pass


class AuthenticationException(CortexException):
    """
    Cortex authentication exception.
    """
    pass


class BadTokenException(CortexException):
    """
    Cortex token invalid exception.
    """
    pass


class ConfigurationException(CortexException):
    """
    Cortex configuration exception.
    """
    pass


class APIException(CortexException):
    """
    Cortex API exception.
    """
    pass
